from pathlib import Path
from typing import Optional, Tuple

import h5py
import numpy as np
import sharedmem
import traceback

from cal_tools.agipdutils import *
from cal_tools.cython import agipdalgs as calgs
from cal_tools.enums import BadPixels, SnowResolution
from cal_tools.tools import get_constant_from_db_and_time
from iCalibrationDB import Constants, Conditions, Detectors


def get_num_cells(fname, loc, module):
    with h5py.File(fname, "r") as f:
        cells = \
            f[f"INSTRUMENT/{loc}/DET/{module}CH0:xtdf/image/cellId"][()]
        if cells.shape[0] == 0:
            return None
        maxcell = np.max(cells)
        options = [4, 32, 64, 76, 128, 176, 202, 250, 352]
        dists = [abs(o - maxcell) for o in options]
        return options[np.argmin(dists)]


def get_acq_rate(slow_paths: Tuple[str, str],
                 fast_paths: Tuple[str, str, int]) -> Optional[float]:
    """Get the acquisition rate from said detector module.

    If the data is available from the middlelayer FPGA_COMP device, then it is
    retrieved from there. If not, the rate is calculated from two different pulses
    time.

    The first entry is deliberatly not used, as the detector just began operating,
    and it might have skipped a train.

    :param slow_paths: in which file and h5 path to look for slow data.
                       The first string is the filename with complete path,
                       the second string is the key `karabo_id_control`

    :param fast_paths: in which module file and h5 path to look for pulses.
                       The first string is the filename with complete path,
                       the second string is the module device name `karabo_id`,
                       the third parameter is the module number, used to navigate
                       through the h5 file structure.

    :return acq_rate: the acquisition rate. If not found in either files, return None.
    """
    # Attempt to look for acquisition rate in slow data
    slow_data_file, karabo_id_control = slow_paths
    slow_data_file = Path(slow_data_file)
    if slow_data_file.exists():  
        slow_data_path = f'CONTROL/{karabo_id_control}/MDL/FPGA_COMP/bunchStructure/repetitionRate/value'  # noqa
        with h5py.File(slow_data_file, "r") as fin:
            if slow_data_path in fin:
                return fin[slow_data_path][3]

    # Compute acquisition rate from fast data
    fast_data_file, karabo_id, module = fast_paths
    fast_data_file = Path(fast_data_file)
    if fast_data_file.exists():
        fast_data_path = f'INSTRUMENT/{karabo_id}/DET/{module}CH0:xtdf/image/pulseId'
        with h5py.File(fast_data_file) as fin:
            if fast_data_path in fin:
                pulses = np.squeeze(fin[fast_data_path][1:3])
                diff = pulses[1] - pulses[0]
                options = {8: 0.5, 4: 1.1, 2: 2.2, 1: 4.5}
                return options.get(diff, None)


def get_gain_setting(fname: str, h5path_ctrl: str) -> int:
    """
    If the data is available from the middlelayer FPGA_COMP device, then it is
    retrieved from there.
    If not, the setting is calculated off `setupr` and `patternTypeIndex`

    gain-setting 1: setupr@dark=8, setupr@slopespc=40
    gain-setting 0: setupr@dark=0, setupr@slopespc=32

    patternTypeIndex 1: High-gian
    patternTypeIndex 2: Medium-gain
    patternTypeIndex 3: Low-gain
    patternTypeIndex 4: SlopesPC

    :param fname: path to file with control information
    :param h5path_ctrl: path to control information inside the file
    :return: gain setting
    """
    gain_path = f'{h5path_ctrl}/gain/value'
    with h5py.File(fname, "r") as fin:
        if gain_path in fin:
            return fin[gain_path][0]
 
        # Get the index at which the train is not zero.
        train_id = fin["INDEX/trainId"][()]
        idx = np.nonzero(train_id)[0][0]

        setupr = fin[f'{h5path_ctrl}/setupr/value'][idx]
        pattern_type_idx = fin[f'{h5path_ctrl}/patternTypeIndex/value'][idx]

        if (setupr == 0 and pattern_type_idx < 4) or (
                setupr == 32 and pattern_type_idx == 4):
            return 0
        elif (setupr == 8 and pattern_type_idx < 4) or (
                setupr == 40 and pattern_type_idx == 4):
            return 1
        else:
            raise ValueError('Could not derive gain setting from setupr and patternTypeIndex')  # noqa


class AgipdCorrections:

    def __init__(self, max_cells, max_pulses,
                 h5_data_path="INSTRUMENT/SPB_DET_AGIPD1M-1/DET/{}CH0:xtdf/",
                 h5_index_path="INDEX/SPB_DET_AGIPD1M-1/DET/{}CH0:xtdf/",
                 corr_bools={}):
        """
        Initialize an AgipdCorrections Class

        :param max_cells: maximum number of memory cells to handle, e.g. if
                         calibration constants only exist for a subset of cells
        :param max_pulses: a range list of pulse indices used for
            calibration and histogram. [start, end, step]
        :param h5_data_path: path in HDF5 file which is prefixed to the
            image/data section
        :param h5_index_path: path in HDF5 file which is prefixed to the
            index section
        :param corr_bools: A dict with all of the correction booleans requested or
                           available

        The following example shows a typical use case:
        .. code-block:: python

            agipd_corr = AgipdCorrections(max_cells, max_pulses,
                                          h5_data_path=h5path,
                                          h5_index_path=h5path_idx,
                                          corr_bools=corr_bools)

            agipd_corr.allocate_constants(modules, (3, mem_cells_db, 512, 128))

            with open(f'{out_folder}/retrieved_constants.yml', "r") as f:
                const_yaml = yaml.load(f.read(), Loader=yaml.FullLoader)

            for mod in modules:
                qm = f"Q{mod // 4 + 1}M{mod % 4 + 1}"
                device = getattr(getattr(Detectors, dinstance), qm)
                agipd_corr.initialize_from_yaml(const_yaml, mod, device)

            data_shape = (n_images_max, 512, 128)
            agipd_corr.allocate_images(data_shape, n_cores_files)

            i_proc=0
            agipd_corr.read_file(i_proc, file_name)

            # Correct first 1000 images
            agipd_corr.correct_agipd(i_proc, first=0, last=1000)

            agipd_corr.write_file(i_proc, file_name, ofile_name)

        """

        # Data description
        self.h5_data_path = h5_data_path
        self.h5_index_path = h5_index_path
        self.rng_pulses = max_pulses
        # avoid list(range(*[0]]))
        self.pulses_lst = list(range(*max_pulses)) \
            if not (len(max_pulses) == 1 and max_pulses[0] == 0) else max_pulses  #noqa
        self.min_pulse = self.pulses_lst[0]
        self.max_pulse = self.pulses_lst[-1]
        self.max_cells = max_cells

        # Correction parameters
        self.baseline_corr_noise_threshold = -1000
        self.snow_resolution = SnowResolution.INTERPOLATE
        self.cm_dark_fraction = 0.66
        self.cm_n_itr = 4
        self.mg_hard_threshold = 100
        self.hg_hard_threshold = 100
        self.noisy_adc_threshold = 0.25

        # Shared variables for data and constants
        self.shared_dict = []
        self.offset = {}
        self.noise = {}
        self.thresholds = {}
        self.frac_high_med = {}
        self.md_additional_offset = {}
        self.rel_gain = {}
        self.mask = {}
        self.xray_cor = {}

        # check if given corr_bools are correct
        tot_corr_bools = ['only_offset', 'adjust_mg_baseline', 'rel_gain',
                          'xray_corr', 'blc_noise', 'blc_hmatch',
                          'blc_stripes', 'blc_set_min', 'match_asics',
                          'corr_asic_diag', 'zero_nans',
                          'zero_orange', 'force_hg_if_below',
                          'force_mg_if_below', 'mask_noisy_adc',
                          'melt_snow', 'common_mode', 'mask_zero_std',
                          'low_medium_gap']

        if set(corr_bools).issubset(tot_corr_bools):
            self.corr_bools = corr_bools
        else:
            bools = list(set(corr_bools) - set(tot_corr_bools))
            raise Exception(f'Correction Booleans: {bools} are not available!')  # noqa

        # Flags allowing for pulse capacitor constant retrieval.
        self.pc_bools = [self.corr_bools.get("rel_gain"),
                         self.corr_bools.get("adjust_mg_baseline"),
                         self.corr_bools.get('blc_noise'),
                         self.corr_bools.get('blc_hmatch'),
                         self.corr_bools.get('blc_stripes'),
                         self.corr_bools.get('melt_snow')]

    def read_file(self, i_proc, file_name):
        """
        Read file with raw data to shared memory

        :param file_name: Name of input file including path
        :param i_proc: Index of shared memory array
        :return:
            - first_index - Index of the first image to be corrected
            - last_index - Index of the last image to be corrected
            - firange - List of images to be corrected

        """
        module_idx = int(file_name.split('/')[-1].split('-')[2][-2:])
        agipd_base = self.h5_data_path.format(module_idx)
        idx_base = self.h5_index_path.format(module_idx)
        data_path = f'{agipd_base}/image'
        data_dict = self.shared_dict[i_proc]

        try:
            f = h5py.File(file_name, 'r')
            group = f[data_path]

            valid, first_index, last_index, valid_trains, valid_indices = \
                self.get_valid_image_idx(idx_base, f)
            firange = self.gen_valid_range(first_index, last_index,
                                           self.max_cells, agipd_base, f,
                                           valid_indices)
            n_img = firange.shape[0]

            data_dict['cellId'][:n_img] = np.squeeze(group['cellId'][firange])
            data_dict['pulseId'][:n_img] = np.squeeze(
                group['pulseId'][firange])
            data_dict['trainId'][:n_img] = np.squeeze(
                group['trainId'][firange])
            data_dict['moduleIdx'][0] = module_idx
            data_dict['nImg'][0] = n_img

            raw_data = group['data'][firange]
            data_dict['data'][:n_img] = raw_data[:, 0]
            data_dict['rawgain'][:n_img] = raw_data[:, 1]

        except Exception as e:
            print(f'Error during reading data from file {file_name}: {e}')
            print(f'Error traceback: {traceback.format_exc()}')
            n_img = 0
            data_dict['nImg'][0] = 0

        return n_img

    def write_file(self, i_proc, file_name, ofile_name):
        """
        Create output file and write corrected data to it

        :param file_name: Name of input file including path
        :param ofile_name: Name of output file including path
        :param i_proc: Index of shared memory array
        """

        module_idx = int(file_name.split('/')[-1].split('-')[2][-2:])
        agipd_base = self.h5_data_path.format(module_idx)
        idx_base = self.h5_index_path.format(module_idx)
        data_path = f'{agipd_base}/image'
        data_dict = self.shared_dict[i_proc]

        with h5py.File(file_name, 'r') as infile:
            with h5py.File(ofile_name, 'w') as outfile:

                n_img = data_dict['nImg'][0]
                if n_img == 0:
                    return
                trains = data_dict['trainId'][:n_img]
                self.copy_and_sanitize_non_cal_data(infile, outfile,
                                                    agipd_base,
                                                    idx_base, trains)

                outfile[data_path]['data'] = data_dict['data'][:n_img]
                outfile[data_path]['gain'] = data_dict['gain'][:n_img]
                outfile[data_path]['blShift'] = data_dict['blShift'][:n_img]
                outfile[data_path]['mask'] = data_dict['mask'][:n_img]
                outfile[data_path]['cellId'] = data_dict['cellId'][:n_img]
                outfile[data_path]['pulseId'] = data_dict['pulseId'][:n_img]
                outfile[data_path]['trainId'] = data_dict['trainId'][:n_img]

    def cm_correction(self, i_proc, asic):
        """
        Perform common-mode correction of data in shared memory

        :param i_proc: Index of shared memory array to process
        :param asic: Asic number to process
        """

        if not self.corr_bools.get("common_mode"):
            return

        fraction = self.cm_dark_fraction
        n_itr = self.cm_n_itr

        cell_id = self.shared_dict[i_proc]['cellId']
        train_id = self.shared_dict[i_proc]['trainId']
        n_img = self.shared_dict[i_proc]['nImg'][0]
        cell_ids = cell_id[train_id == train_id[0]]
        n_cells = cell_ids.size
        data = self.shared_dict[i_proc]['data'][:n_img].reshape(-1, n_cells, 8,
                                                                64, 2, 64)

        # Loop over iterations
        for i in range(n_itr):
            # Loop over rows of cells
            first = 0
            for cell_row in range(11):
                last = first + cell_ids[(cell_ids // 32) == cell_row].size
                if first == last:
                    continue
                asic_data = data[:, first:last, asic % 8, :, asic // 8, :]

                # Cell common mode
                cell_cm_sum, cell_cm_count = \
                    calgs.sum_and_count_in_range_cell(asic_data, -25., 25.)
                cell_cm = cell_cm_sum / cell_cm_count

                cell_cm[cell_cm_count < fraction * 32 * 256] = 0
                asic_data[...] -= cell_cm[None, None, :, :]

                # Asics common mode
                asic_cm_sum, asic_cm_count = \
                    calgs.sum_and_count_in_range_asic(asic_data, -25., 25.)
                asic_cm = asic_cm_sum / asic_cm_count

                asic_cm[asic_cm_count < fraction * 64 * 64] = 0
                asic_data[...] -= asic_cm[:, :, None, None]

                first = last

    def mask_zero_std(self, i_proc, cells):
        """
        Add bad pixel bit: DATA_STD_IS_ZERO to the mask of bad pixels

        Pixel is bad if standard deviation for a given pixel and
        given memory cell is zero

        :param i_proc: Index of shared memory array to process
        :param cells: List of cells to be considered
        """
        if not self.corr_bools.get("mask_zero_std"):
            return

        module_idx = self.shared_dict[i_proc]['moduleIdx'][0]
        n_img = self.shared_dict[i_proc]['nImg'][0]
        data = self.shared_dict[i_proc]['data'][:n_img]
        cellid = self.shared_dict[i_proc]['cellId'][:n_img]
        mask = self.mask[module_idx]  # shape of n_cells, x, y

        for c in cells:
            std = np.nanstd(data[cellid == c, ...], axis=0)
            mask[:, c, std == 0] |= BadPixels.DATA_STD_IS_ZERO.value

    def correct_agipd(self, i_proc, first, last):
        """
        Perform image-wise correction of data in shared memory

        :param first: Index of the first image to be corrected
        :param last: Index of the last image to be corrected
        :param i_proc: Index of shared memory array to process
        """

        module_idx = self.shared_dict[i_proc]['moduleIdx'][0]
        data = self.shared_dict[i_proc]['data'][first:last]
        bl_shift = self.shared_dict[i_proc]['blShift'][first:last]
        rawgain = self.shared_dict[i_proc]['rawgain'][first:last]
        gain = self.shared_dict[i_proc]['gain'][first:last]
        cellid = self.shared_dict[i_proc]['cellId'][first:last]
        mask = self.shared_dict[i_proc]['mask'][first:last]

        # first evaluate the gain into 0, 1, 2 --> high, medium, low
        t0 = self.thresholds[module_idx][0]
        t1 = self.thresholds[module_idx][1]

        rgain = None
        raw_data = None
        if self.corr_bools.get('melt_snow'):
            rgain = rawgain / t0[cellid, ...]
            raw_data = np.copy(data)

        # Often most pixels are in high-gain, so it's more efficient to
        # set the whole output block to zero than select the right pixels.
        gain[:] = 0
        # exceeding first threshold means data is medium or low gain
        gain[rawgain > t0[cellid, ...]] = 1
        # exceeding also second threshold means data is low gain
        gain[rawgain > t1[cellid, ...]] = 2

        offsetb = self.offset[module_idx][:, cellid]

        # force into high or medium gain if requested
        if self.corr_bools.get('force_mg_if_below'):
            gain[(gain == 2) & (
                        (data - offsetb[1]) < self.mg_hard_threshold)] = 1

        if self.corr_bools.get('force_hg_if_below'):
            gain[(gain > 0) & (
                        (data - offsetb[0]) < self.hg_hard_threshold)] = 0

        # choose constants according to gain setting
        off = calgs.gain_choose(gain, offsetb)
        del offsetb

        tmask = self.mask[module_idx][:, cellid]
        msk = calgs.gain_choose_int(gain, tmask)
        del tmask

        # same for relative gain and then bad pixel mask
        if hasattr(self, "rel_gain"):
            # get the correct rel_gain depending on cell-id
            rc = self.rel_gain[module_idx][:, cellid]
            rel_cor = calgs.gain_choose(gain, rc)
            del rc

        # subtract offset
        data -= off
        del off

        # before doing relative gain correction we need to evaluate any
        # baseline shifts
        # as they are effectively and additional offset in the data
        if (self.corr_bools.get('blc_noise') or
                self.corr_bools.get('blc_hmatch') or
                self.corr_bools.get('blc_stripes')):

            # do this image wise, as the shift is per image
            for i in range(data.shape[0]):

                # first correction requested may be to evaluate shift via
                # noise peak
                if self.corr_bools.get('blc_noise'):
                    mn_noise = np.nanmean(self.noise[module_idx][0, cellid[i]])  #noqa
                    dd, sh = baseline_correct_via_noise(data[i],
                                                        mn_noise,
                                                        gain[i],
                                                        self.baseline_corr_noise_threshold)  #noqa
                # if not we continue with initial data
                else:
                    dd = data[i]

                # if we have enough pixels in medium or low gain and
                # correction via hist matching is requested to this now
                gcrit = np.count_nonzero(gain[i] > 0) > 1000
                if (gcrit and self.corr_bools.get('blc_hmatch') and
                        hasattr(self, "rel_gain")):
                    dd2, sh2 = correct_baseline_via_hist(data[i],
                                                         rel_cor[i],
                                                         gain[i])
                    data[i] = np.maximum(dd, dd2)
                    sh = np.minimum(sh, sh2)
                    # finally correct diagonal effects if requested
                    if self.corr_bools.get('corr_asic_diag'):
                        ii = data[i, ...]
                        gg = gain[i, ...]
                        adim = correct_baseline_via_hist_asic(ii, gg)
                        data[i, ...] = adim
                # if there is not enough medium or low gain data to do an
                # evaluation, do nothing
                else:
                    data[i, ...] = dd
                    bl_shift[i] = sh

                if self.corr_bools.get('blc_stripes'):
                    fmh = self.frac_high_med[module_idx][cellid[i]]
                    dd, sh = baseline_correct_via_stripe(data[i, ...],
                                                         gain[i, ...],
                                                         msk[i, ...],
                                                         fmh)
                    data[i, ...] = dd
                    bl_shift[i] = sh

        # Correct for relative gain
        if self.corr_bools.get("rel_gain") and hasattr(self, "rel_gain"):
            data *= rel_cor
            del rel_cor

        # Set negative values for medium gain to 0
        if self.corr_bools.get('blc_set_min'):
            data[(data < 0) & (gain == 1)] = 0

        # Adjust medium gain baseline to match highest high gain value
        if self.corr_bools.get("adjust_mg_baseline"):
            mgbc = self.md_additional_offset[module_idx][cellid, ...]
            data[gain == 1] += mgbc[gain == 1]
            del mgbc

        # Do xray correction if requested
        if self.corr_bools.get("xray_corr"):
            data *= self.xray_cor[module_idx]

        if self.corr_bools.get('melt_snow'):
            melt_snowy_pixels(raw_data, data, gain, rgain,
                              self.snow_resolution)
            del raw_data
            del rgain

        # Inner ASIC borders are matched to the same signal level
        if self.corr_bools.get("match_asics"):
            data = match_asic_borders(data, 8, module_idx)

        # Add any non-finite values to the mask, zero them
        if self.corr_bools.get("zero_nans"):
            bidx = ~np.isfinite(data)
            data[bidx] = 0
            msk[bidx] |= BadPixels.VALUE_IS_NAN.value
            del bidx

        # Add pixels with unrealistically high and low values to the mask.
        # Zero them.
        if self.corr_bools.get("zero_orange"):
            bidx = (data < -1e7) | (data > 1e7)
            data[bidx] = 0
            msk[bidx] |= BadPixels.VALUE_OUT_OF_RANGE.value
            del bidx

        # Mask entire ADC if they are noise above a threshold
        if self.corr_bools.get("mask_noisy_adc"):
            make_noisy_adc_mask(msk, self.noisy_adc_threshold)

        mask[...] = msk[...]

    def get_valid_image_idx(self, idx_base, infile, index_v=2):
        """ Return the indices of valid data
        """

        if index_v == 2:
            count = np.squeeze(infile[idx_base + "image/count"])
            first = np.squeeze(infile[idx_base + "image/first"])
            if np.count_nonzero(count != 0) == 0:
                raise IOError("File has no valid counts")
            valid = count != 0
            idxtrains = np.squeeze(infile["/INDEX/trainId"])
            medianTrain = np.nanmedian(idxtrains)
            lowok = (idxtrains > medianTrain - 1e4)
            highok = (idxtrains < medianTrain + 1e4)
            valid &= lowok & highok

            last_index = int(first[valid][-1] + count[valid][-1])
            first_index = int(first[valid][0])

            # do actual validity filtering:
            validc, validf = count[valid], first[valid]
            valid_indices = np.concatenate([np.arange(validf[i],
                                                      validf[i]+validc[i])
                                            for i in range(validf.size)],
                                            axis=0)
            valid_indices = np.squeeze(valid_indices).astype(np.int32)

        elif index_v == 1:
            status = np.squeeze(infile[idx_base + "image/status"])
            if np.count_nonzero(status != 0) == 0:
                raise IOError("File {} has no valid counts".format(
                    infile))
            last = np.squeeze(infile[idx_base + "image/last"])
            first = np.squeeze(infile[idx_base + "image/first"])
            valid = status != 0
            last_index = int(last[status != 0][-1]) + 1
            first_index = int(first[status != 0][0])

            idxtrains = np.squeeze(infile["/INDEX/trainId"])
            medianTrain = np.nanmedian(idxtrains)
            lowok = (idxtrains > medianTrain - 1e4)
            highok = (idxtrains < medianTrain + 1e4)
            valid &= lowok & highok
            valid_indices = None
        else:
            raise AttributeError(
                "Not a known raw format version: {}".format(index_v))

        return valid, first_index, last_index, idxtrains, valid_indices

    def gen_valid_range(self, first_index, last_index, max_cells, agipd_base,
                        infile, valid_indices=None):
        """ Generate an index range to pass to `correctAGIPD`.
        """

        allcells = np.squeeze(infile[agipd_base + "image/cellId"])
        allpulses = np.squeeze(infile[agipd_base + "image/pulseId"])
        if valid_indices is not None:
            allcells = allcells[valid_indices]
            allpulses = allpulses[valid_indices]
        else:
            allcells = allcells[first_index:last_index]
            allpulses = allpulses[first_index:last_index]

        # Calculate the pulse step from the chosen max_pulse range
        if len(self.rng_pulses) == 3:
            pulse_step = self.rng_pulses[2]
        else:
            pulse_step = 1

        # Make sure the range max doesn't have non-valid idx.
        if self.pulses_lst[-1] + pulse_step > last_index-1:
            last_pulse = last_index-1
        else:
            last_pulse = self.pulses_lst[-1] + pulse_step
        # Check if 1st pulse idx was out of valid range.
        # If that is the case only calibrate the last step pulse.
        if self.pulses_lst[0] >= last_pulse:
            first_pulse = last_pulse - pulse_step
        else:
            first_pulse = self.pulses_lst[0]

        # collect the pulses to be calibrated
        cal_pulses = allpulses[first_pulse:last_pulse:pulse_step]

        # These parameters required also for histogram pulse range.
        self.min_pulse = np.min(cal_pulses)
        self.max_pulse = np.max(cal_pulses)

        can_calibrate = (allcells < max_cells)

        # Check if a specific pulses need to be calibrated
        # or with only simple pulse ranging.
        if pulse_step > 1 or \
                allpulses[first_pulse] >= allpulses[last_pulse]:
            can_calibrate &= np.isin(allpulses, cal_pulses)
        else:
            can_calibrate &= (allpulses <= self.max_pulse)
            can_calibrate &= (allpulses >= self.min_pulse)

        if np.count_nonzero(can_calibrate) == 0:
            return
        if valid_indices is None:
            firange = np.arange(first_index, last_index)
        else:
            firange = valid_indices
        firange = firange[can_calibrate]

        return firange

    def copy_and_sanitize_non_cal_data(self, infile, outfile, agipd_base,
                                       idx_base, trains):
        """ Copy and sanitize data in `infile` that is not touched by
        `correctAGIPD`
        """

        # these are touched in the correct function, do not copy them here
        dont_copy = ["data", "cellId", "trainId", "pulseId", "status",
                     "length"]
        dont_copy = [agipd_base + "image/{}".format(do)
                     for do in dont_copy]

        # don't copy these as we may need to adjust if we filter trains
        dont_copy += [idx_base + "{}/first".format(do)
                      for do in ["image", ]]
        dont_copy += [idx_base + "{}/count".format(do)
                      for do in ["image", ]]
        dont_copy += [idx_base + "{}/last".format(do)
                      for do in ["image", ]]
        dont_copy += [idx_base + "{}/status".format(do)
                      for do in ["image", ]]

        # a visitor to copy everything else
        def visitor(k, item):
            if k not in dont_copy:

                if isinstance(item, h5py.Group):
                    outfile.create_group(k)
                elif isinstance(item, h5py.Dataset):
                    group = str(k).split("/")
                    group = "/".join(group[:-1])
                    infile.copy(k, outfile[group])

        infile.visititems(visitor)

        # sanitize indices
        for do in ["image", ]:
            # uq: INDEX/trainID
            # fidxv: INDEX/.../image/first idx values
            # cntsv: INDEX/.../image/counts values

            # Extract parameters through identifying
            # unique trains, index and numbers.
            uq, fidxv, cntsv = np.unique(trains, return_index=True,
                                         return_counts=True)

            # Validate calculated CORR INDEX contents by checking difference between
            # trainId stored in RAW data and trains from
            train_diff = np.isin(np.array(infile["/INDEX/trainId"]), uq,
                                 invert=True)

            # Insert zeros for missing trains.
            # fidxv and cntsv should have same length as
            # raw INDEX/.../image/first and INDEX/.../image/count,
            # respectively

            # first_inc = first incrementation
            first_inc = True
            for i, diff in enumerate(train_diff):
                if diff:
                    if i < len(cntsv):
                        cntsv = np.insert(cntsv, i, 0)
                        if i == 0:
                            fidxv = np.insert(fidxv, i, 0)
                        else:
                            fidxv = np.insert(fidxv, i, fidxv[i])
                    else:
                        # append if at the end of the array
                        cntsv = np.append(cntsv, 0)
                        # increment fidxv once with the
                        # no. of processed mem-cells.
                        if first_inc:
                            fidxv = np.append(fidxv,
                                              (2 * fidxv[i-1]) - fidxv[i-2])
                            first_inc = False
                        else:
                            fidxv = np.append(fidxv, fidxv[i-1])

            # save INDEX contents (first, count) in CORR files
            outfile.create_dataset(idx_base + "{}/first".format(do),
                                        fidxv.shape,
                                        dtype=fidxv.dtype,
                                        data=fidxv,
                                        fletcher32=True)
            outfile.create_dataset(idx_base + "{}/count".format(do),
                                        cntsv.shape,
                                        dtype=cntsv.dtype,
                                        data=cntsv,
                                        fletcher32=True)

    def retrieve_constant_and_time(self, const_dict, device,
                                   cal_db_interface, creation_time):
        """
        A wrapper around get_constant_from_db_and_time to get
        constant image and creation time using a dictionary of
        the names of the required constant from the data base
        and their parameter operating conditions.

        :param const_dict: (Dict) A dictionary with constants to retrieve
        and their operating conditions.
        e.g.{ "Offset": ["zeros", (128, 512, memory_cells, 3), "Dark"]
              "Dark-cond": {Condition-name}: condition-value }
        :param device: (iCalibrationDB obj) Object for AGIPD's device
        :param cal_db_interface: (Str) The port interface for
        calibrationDBRemote
        :param creation_time: (Str) Raw data creation time
        :return:
            cons_data: (np.array) Constant data
            when: (Dict) A dictionary with all retrieved constants
            and their creation-time. {"Constant-Name": "creation-time"}
        """
        when = {}
        cons_data = {}
        for cname, cval in const_dict.items():
            condition = \
                getattr(Conditions, cval[2][0]).AGIPD(**cval[2][1])
            cons_data[cname], when[cname] = \
                get_constant_from_db_and_time(device,
                                              getattr(Constants.AGIPD, cname)(),  # noqa
                                              condition,
                                              getattr(np, cval[0])(cval[1]),
                                              cal_db_interface,
                                              creation_time,
                                              print_once=0)
        return cons_data, when

    def init_constants(self, cons_data, module_idx):
        """
        For CI derived gain, a mean multiplication factor of 4.48 compared
        to medium gain is used, as no reliable CI data for all memory cells
        exists of the current AGIPD instances.

        Relative gain is derived both from pulse capacitor as well as flat
        field data:

        * from the pulse capacitor data we get the relative slopes of a
          given pixel's memory cells with respect to all memory cells of that
          pixel:

             rpch = m_h / median(m_h)

          where m_h is the high gain slope m of each memory cell of the pixel.

        * we also derive the factor between high and medium gain in a
          similar way and scale it to be relative to the pixels memory cells:

             fpc = m_m/m_h
             rfpc = fpc/ median(fpc)

          where m_m is the medium gain slope of all memory cells of a given
          pixel and m_h is the high gain slope as above

        * finally, we get the relative X-ray gain of all memory cells for a
          given pixel from flat field data:

             ff = median(m_ff)
             ff /= median(ff)

          where m_ff is the flat field derived (high gain) slope of all
          memory cells of a given pixel. The pixel values are then scaled to
          the complete module_idx. Note that the first 32 memory cells are known
          to exhibit differing behaviour and are skipped if possible.

        With this data the relative gain for the three gain stages evaluates
        to:

            high gain = ff * rpch
            medium gain = ff * rfpc
            low gain = medium gain / 4.48

        :param cons_data: A dictionary for each retrieved constant value.
        :param module_idx: A module_idx index
        :return:
        """

        self.offset[module_idx][...] = cons_data["Offset"].transpose()[...]
        self.noise[module_idx][...] = cons_data["Noise"].transpose()[...]
        self.thresholds[module_idx][...] = cons_data["ThresholdsDark"].transpose()[:3,...]  # noqa

        if self.corr_bools.get("low_medium_gap"):
            t0 = self.thresholds[module_idx][0]
            t1 = self.thresholds[module_idx][1]
            t1[t1 <= 1.05 * t0] = np.iinfo(np.uint16).max

        bpixels = cons_data["BadPixelsDark"].astype(np.uint32)

        if self.corr_bools.get("xray_corr"):
            bpixels |= cons_data["BadPixelsFF"].astype(np.uint32)[..., :bpixels.shape[2], None]  # noqa
            slopesFF = np.squeeze(cons_data["SlopesFF"])

            if len(slopesFF.shape) == 4:
                slopesFF = slopesFF[..., 0]

            # first 32 cells are known to behave differently so if we can avoid
            # them
            # when calculating the mean X-ray derived gain slope for each pixel
            if slopesFF.shape[2] > 32:
                xray_cor = np.nanmedian(
                    slopesFF[..., 32:min(96, self.max_cells)], axis=2)
            elif slopesFF.shape[2] > 2:
                xray_cor = np.nanmedian(
                    slopesFF[..., :min(96, self.max_cells)], axis=2)
            else:
                xray_cor = np.squeeze(slopesFF[..., 0])

            # relative X-ray correction is normalized by the median of all pixels
            xray_cor /= np.nanmedian(xray_cor)
            self.xray_cor[module_idx][...] = xray_cor.transpose()[...]

        # add additional bad pixel information
        if any(self.pc_bools):
            bppc = np.moveaxis(cons_data["BadPixelsPC"].astype(np.uint32), 0, 2)  # noqa
            bpixels |= bppc[..., :bpixels.shape[2], None]

            slopesPC = cons_data["SlopesPC"].astype(np.float32)

            # this will handle some historical data in a different format
            # constant dimension injected first
            if slopesPC.shape[0] == 10 or slopesPC.shape[0] == 11:
                slopesPC = np.moveaxis(slopesPC, 0, 3)
                slopesPC = np.moveaxis(slopesPC, 0, 2)

            # calculate relative gain from the constants
            rel_gain = np.ones((128, 512, self.max_cells, 3), np.float32)
            # high gain slope from pulse capacitor data
            pc_high_m = slopesPC[..., :self.max_cells, 0]
            pc_high_l = slopesPC[..., :self.max_cells, 1]
            # medium gain slope from pulse capacitor data
            pc_med_m = slopesPC[..., :self.max_cells, 3]
            pc_med_l = slopesPC[..., :self.max_cells, 4]

            # calculate ratio high to medium gain over memory cells
            pc_high_ave = np.nanmean(pc_high_m, axis=(0,1))
            pc_med_ave = np.nanmean(pc_med_m, axis=(0,1))
            frac_high_med = pc_high_ave / pc_med_ave

            # calculate additional medium-gain offset
            md_additional_offset = pc_high_l - pc_med_l * pc_high_m / pc_med_m

            # Calculate relative gain
            rel_gain[..., 0] = pc_high_m / pc_high_ave
            rel_gain[..., 1] = pc_med_m / pc_med_ave * frac_high_med
            rel_gain[..., 2] = rel_gain[..., 1] * 4.48

            self.md_additional_offset[module_idx][...] = md_additional_offset.transpose()[...]  # noqa
            self.rel_gain[module_idx][...] = rel_gain[...].transpose()
            self.frac_high_med[module_idx][...] = frac_high_med

        self.mask[module_idx][...] = bpixels.transpose()[...]

        return

    def initialize_from_yaml(self, const_yaml, module_idx, device):
        """
        Initialize calibration constants from a yaml file

        :param const_yaml: (Dict) fromed from a yaml file in pre-notebook,
        which consists of metadata of either the constant file path or the
        empty constant shape, and the creation-time of the retrieved constants
        :param module_idx: Index of module
        :return:
        """

        # string of the device name.
        dname = device.device_name
        cons_data = dict()
        when = dict()
        for cname, mdata in const_yaml[dname].items():
            when[cname] = mdata["creation-time"]
            if when[cname]:
                cf = h5py.File(mdata["file-path"], "r")
                cons_data[cname] = np.copy(cf[f"{dname}/{cname}/0/data"])
                cf.close()
            else:
                # Create empty constant using the list elements
                cons_data[cname] = \
                    getattr(np, mdata["file-path"][0])(mdata["file-path"][1])

        self.init_constants(cons_data, module_idx)

        return when

    def initialize_from_db(self, cal_db_interface, creation_time, memory_cells,
                           bias_voltage, photon_energy, gain_setting,
                           acquisition_rate, module_idx, device, only_dark=False):
        """ Initialize calibration constants from the calibration database

        :param dbparms: a tuple containing relevant database parameters,
        can be either:
            * cal_db_interface, creation_time, max_cells_db, bias_voltage,
              photon_energy in which case the db timeout is set to 300 seconds,
              the cells to query dark image derived constants from the
              database is set to the global value

            * cal_db_interface, creation_time, max_cells_db, bias_voltage,
              photon_energy, max_cells_db_dark
              additionally a timeout is given

        :param qm: quadrant and module of the constants to load in Q1M1
        notation
        :param only_dark: load only dark image derived constants. This
            implies that a `calfile` is used to load the remaining
            constants. Useful to reduce DB traffic and interactions
            for non-frequently changing constants, i.e. such which are
            not usually updated during a beamtime.

        The `cal_db_interface` parameter in the `dbparms` tuple may be in
        one of the following notations:
            * tcp://host:port to directly identify the host and port to
              connect to
            * tcp://host:port_low#port_high to specify a port range from
              which a random port will be picked. E.g. specifying

              tcp://max-exfl016:8015#8025

              will randomly pick an address in the range max-exfl016:8015 and
              max-exfl016:8025.

        The latter notation allows for load-balancing.

        This routine loads the following constants as given in
        `iCalibrationDB`:

            Dark Image Derived
            ------------------

            * Constants.AGIPD.Offset
            * Constants.AGIPD.Noise
            * Constants.AGIPD.BadPixelsDark
            * Constants.AGIPD.ThresholdsDark

            Pulse Capacitor Derived
            -----------------------

            * Constants.AGIPD.SlopesPC

            Flat-Field Derived

            * Constants.AGIPD.SlopesFF

        """

        const_dict = \
            assemble_constant_dict(self.corr_bools, self.pc_bools,
                                   memory_cells, bias_voltage, gain_setting,
                                   acquisition_rate, photon_energy,
                                   beam_energy=None, only_dark=only_dark)

        cons_data, when = \
            self.retrieve_constant_and_time(const_dict, device,
                                            cal_db_interface,
                                            creation_time)

        self.init_constants(cons_data, module_idx)

        return when

    def allocate_constants(self, modules, constant_shape):
        """
        Allocate memory for correction constants

        :param modules: Module indises
        :param constant_shape: Shape of expected constants (gain, cells, x, y)
        """
        for module_idx in modules:
            self.offset[module_idx] = sharedmem.empty(constant_shape,
                                                      dtype='f4')
            self.thresholds[module_idx] = sharedmem.empty(constant_shape,
                                                          dtype='f4')
            self.noise[module_idx] = sharedmem.empty(constant_shape,
                                                     dtype='f4')

            self.md_additional_offset[module_idx] = sharedmem.empty(
                constant_shape[1:], dtype='f4')
            self.rel_gain[module_idx] = sharedmem.empty(constant_shape,
                                                        dtype='f4')
            self.frac_high_med[module_idx] = sharedmem.empty(constant_shape[1],
                                                             dtype='f4')

            self.mask[module_idx] = sharedmem.empty(constant_shape, dtype='i4')
            self.xray_cor[module_idx] = sharedmem.empty(constant_shape[2:],
                                                        dtype='f4')

    def allocate_images(self, shape, n_cores_files):
        """
        Allocate memory for image data

        :param shape: Shape of expected data (nImg, x, y)
        :param n_cores_files: Number of files, handled in parallel
        """

        self.shared_dict = []
        for i in range(n_cores_files):
            self.shared_dict.append({})
            self.shared_dict[i]['cellId'] = sharedmem.empty(shape[0],
                                                            dtype='i4')
            self.shared_dict[i]['pulseId'] = sharedmem.empty(shape[0],
                                                             dtype='i4')
            self.shared_dict[i]['trainId'] = sharedmem.empty(shape[0],
                                                             dtype='i4')
            self.shared_dict[i]['moduleIdx'] = sharedmem.empty(1, dtype='i4')
            self.shared_dict[i]['nImg'] = sharedmem.empty(1, dtype='i4')
            self.shared_dict[i]['mask'] = sharedmem.empty(shape, dtype='i4')
            self.shared_dict[i]['data'] = sharedmem.empty(shape, dtype='f4')
            self.shared_dict[i]['rawgain'] = sharedmem.empty(shape, dtype='u2')
            self.shared_dict[i]['gain'] = sharedmem.empty(shape, dtype='u1')
            self.shared_dict[i]['blShift'] = sharedmem.empty(shape[0],
                                                             dtype='f4')
